/*
 *	%Z% %M% %I% %D% %Q%
 *
 *  Copyright (C) The Santa Cruz Operation, Inc. 1983
 *
 *  Function:	dialer program for the Hayes Smartmodem 1200
 *
 *  Usage:	dial ttyname telnumber speed 
 *
 *  Returns:	 0 if a connection was made.
 *		-1 otherwise, logs reason for failure.
 *
 *  Support:	logent.o ulockf.o
 *		Must run setuid uucp to make log entries.
 *
 *  Documents:	Smartmodem 1200/1200b manual
 *
 */

#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <termio.h>
#include <pwd.h>

static char Sccsid[] = "%Z% %M% %I% %D% %Q%";


#define SAME		0
#define	SUCCESS		0
#define FAIL		-1
#define MAXLINE		80
#define LOGDIR		"/usr/spool/uucp"

#ifndef NODEBUG
#define DEBUG(l, f, s)	if (Debug >= l) fprintf(stderr, f, s)
#else
#define DEBUG(l, f, s)
#endif

/*
 *  MDSETUP - Speaker on, full duplex, no echo, enable result codes,
 *		send result codes as words, tone dial.
 */
#define	MDSETUP		"ATM1E0Q0V1T\r"

/*
 *  MDVALID - Allow only these characters to reach the modem.
 */
#define MDVALID		"0123456789TP,*#()-"

/*
 *  MDESCAPE - Takes modem out of online state to accept commands.
 */
#define	MDESCAPE	"+++"

/*
 *  MDHANGUP - Force modem to drop carrier.
 */
#define	MDHANGUP	"ATH\r"

/*
 *  MDRESET - Reset modem to default power-up state.
 */
#define	MDRESET		"ATZ\r" 

/*
 *  These defines are used to determine how long the dialer timeout
 *  should be.  MDPULSDLY can be changed, but MDPAUSDLY requires
 *  reprogramming modem register S8 to be effective.
 */
#define	MDPULSCHR	'P'
#define	MDPULSDLY	15

#define	MDPAUSCHR	','
#define	MDPAUSDLY	2

/*
 *  Possible messages produced by modem.
 *  We ignore RING since the HAYES only produces it when answering.
 */
#define	OK		0
#define	CONNECT		1
#define	NOCARRIER	2
#define	ERROR		3

char *mdmsgs[] = {
/* 0 */	"OK",
/* 1 */	"CONNECT",
/* 2 */	"NO CARRIER",
/* 3 */	"ERROR",
	0
};

char *strchr();
int alrmint();
struct termio term;
int Debug;				/* set when debug flag is given	*/
int dialing;				/* set while modem is dialing	*/
int fd;					/* file descriptor for acu	*/
char User[10] = "unknown";		/* user id for logfile entries	*/
char Rmtname[] = "unknown";		/* remote system for logfile	*/


main(argc,argv)
	int argc;
	char *argv[];
{
	char *acu;			/* device to dial through	*/
	char *phone;			/* phone number to dial		*/
	char command[MAXLINE];		/* modem command buffer		*/
	int baudrate;			/* baud rate of modem		*/
	int errflag = 0;		/* set on errors		*/
	int hflag = 0;			/* set to hangup modem		*/
	int timeout;			/* how long to wait for alarm	*/
	extern int optind;
	extern char *optarg;
	int c;
	char *p;
	struct passwd *pwd, *getpwuid();

	/*
	 *  Find out who we are for the logfile
	 *  Chdir to LOGDIR so we can create tmp files.
	 */
	if ((pwd = getpwuid(getuid())) != NULL)
		strncpy(User, pwd->pw_name, 10);

	if (chdir(LOGDIR) < 0) {
		fprintf(stderr, "dial: cannot chdir to %s\n", LOGDIR);
		logent(LOGDIR, "CHDIR FAILED"); /* try anyway */
		exit(FAIL);
	}

	while ((c = getopt(argc, argv, "hx:")) != EOF)
		switch(c) {
			case 'h':
				hflag++;
				break;
			case 'x':
				Debug = atoi(optarg);
				break;
			case '?':
				errflag++;
				break;
		}
	
	if (errflag || argc - optind != 3) {
		fprintf(stderr,"Usage: dial devicename number speed\n");
		logent("INVALID ARGUMENTS", "ACU FAILED");
		exit(FAIL);
	}

	acu = argv[optind++];
	phone = argv[optind++];
	
	if (strlen(phone) != strspn(phone, MDVALID)) {
		logent("INVALID PHONE NUMBER", "ACU FAILED");
		fprintf(stderr, "dial: Bad phone number %s\n", phone);
		exit(FAIL);
	}

	switch(atoi(argv[optind])) {
		case 300:
			baudrate = B300;
			break;
		case 1200:
			baudrate = B1200;
			break;
		default:
			fprintf(stderr, "dial: Bad speed: %s\n", argv[optind]);
			logent("INVALID SPEED", "ACU FAILED");
			exit(FAIL);
	}		
	/*
	 *  Must open with O_NDELAY set or the open may hang.
	 */
	if ((fd = open(acu, O_RDWR | O_NDELAY)) < 0) {
		fprintf(stderr, "dial: Can't open device: %s\n", acu);
		logent("CANNOT OPEN DEVICE", "ACU FAILED");
		exit(FAIL);
	}
	/*
	 * set line for no echo and correct speed
	 */
	errflag = ioctl(fd, TCGETA, &term);
	term.c_cflag &= ~CBAUD;
	term.c_cflag |= CLOCAL | HUPCL | baudrate;
	term.c_lflag &= ~ECHO;
	term.c_cc[VMIN] = '\1';
	term.c_cc[VTIME] = '\0';
	errflag = ioctl(fd, TCSETA, &term);
	if (errflag) {
		DEBUG(1, "dial: ioctl error on %s\n", acu);
		logent("IOCTL ERROR", "ACU FAILED");
		cleanup(FAIL);
	}
	fcntl(fd, F_SETFL, fcntl(fd, F_GETFL, 0) & ~O_NDELAY);


	/*
	 *  Timeout after 5 seconds if no response
	 */
	timeout = 5;

	/*
	 * Hangup and exit if requested
	 */
	if (hflag)
		cleanup(hangup(timeout));

	/*
	 *  Reset and put modem into command mode
	 */
reset:
	if (mdwrite(MDRESET) == FAIL)
		cleanup(FAIL);
	if (mdread(timeout) != OK) {
		if (hangup(timeout) == FAIL)
			cleanup(FAIL);
		goto reset;
	}

	/*
	 *  Must wait at least 0.5 seconds after reset
	 *  for HAYES to recover and receive commands.
	 */
	sleep(1);

	/*
	 *  Initialize the modem
	 */
	if (mdwrite(MDSETUP) == FAIL)
		cleanup(FAIL);
	if (mdread(timeout) != OK) {
		if (hangup(timeout) == FAIL)
			cleanup(FAIL);
		goto reset;
	}

	/*
	 *  Build up the phone number
	 */
	sprintf(command, "ATD%s\r", phone);

	/*
	 *  Set up a timeout for the connect.
	 *    Add in MDPAUSDLY seconds more for each pause character
	 *    Pulse dialing takes MDPULSDLY seconds longer too
	 */
	timeout = 6 * strlen(phone) + 15;
	for (p = phone; (p = strchr(p, MDPAUSCHR)) != NULL; p++)
		timeout += MDPAUSDLY;
	if (strchr(phone, MDPULSCHR) != NULL)
		timeout += MDPULSDLY;
	if (timeout < 30)
		timeout = 30;

redial:
	if (mdwrite(command) == FAIL)
		cleanup(FAIL);

	dialing = 1;
	DEBUG(6, "wait for connect - timeout %d\n", timeout);

	switch (mdread(timeout)) {
		case OK:
		case ERROR:
			goto redial;
		case CONNECT:
			cleanup(SUCCESS);
		case NOCARRIER:
			logent("NO CARRIER", "ACU FAILED");
			cleanup(FAIL);
		default:
			cleanup(FAIL);
	}
}


/*
 *  hangup(htime)
 *
 *  Function:	Forces the modem to drop carrier and hang up the phone.
 *		Reads are allowed htime seconds before timeout.
 *
 *  Returns:	SUCCESS if disconnect completed.
 *		FAIL on failure, phone may still be connected.
 *
 */
hangup(htime)
	int htime;
{
	DEBUG(4, "hangup - timeout %d\n", htime);
	sleep(1);
	if (mdwrite(MDESCAPE) == FAIL)
		return(FAIL);
	sleep(1);
	/*
	 * MDESCAPE will return OK only if online, so ignore error return
	 */
	mdread(htime);
	if (mdwrite(MDHANGUP) == FAIL)
		return(FAIL);
	if (mdread(htime) != OK)
		return(FAIL);
	return(SUCCESS);
}
	
/*
 *  mdread(rtime)
 *
 *  Function:	Reads from the ACU until it finds a valid response (found
 *		in mdmsgs) or times out after rtime seconds.
 *
 *  Returns:	The index in mdmsgs of the modem response found.
 *		FAIL on timeout.
 *
 */
mdread(rtime)
	int rtime;
{
	char c, **mp;
	register char *bp;
	char buf[MAXLINE];

	bp = buf;
	signal(SIGALRM, alrmint);
	alarm(rtime);
	DEBUG(6, "MODEM returned %s", "<<");
	while (read(fd, &c, 1) == 1) {
		if ((*bp = c) != '\0')
			*++bp = '\0';
		DEBUG(6, (c >= 040 ? "%c" : "^%c"), (c >= 040 ? c : c+0100));
		if (bp >= buf + MAXLINE) {
			alarm(0);
			DEBUG(4,">>-%s\n","FAIL");
			return(FAIL);
		}
		for (mp = mdmsgs; *mp; ++mp)
			if (substr(*mp, buf) == SUCCESS) {
				alarm(0);
				DEBUG(6,">>-%s\n", "OK");
				DEBUG(4,"got %s\n", mdmsgs[mp - mdmsgs]);
				return(mp - mdmsgs);
			}
	}
	alarm(0);
	DEBUG(6,">>-%s","FAIL");
	DEBUG(4, "acu lost line c=%c\n", c);
	logent("LOST LINE", "ACU FAILED");
	return(FAIL);
}


/*
 *  mdwrite(c)
 *
 *  Function:	Outputs the string pointed to by c to the ACU device.
 *
 *  Returns:	SUCCESS on completion.
 *		FAIL on write errors.
 *
 */
mdwrite(c)
	register char *c;
{
	/*
	 *  Give modem a chance to recover before writing.
	 */
	sleep(1);
	DEBUG(6, "Sent MODEM %s", "<<");
	while (*c) {
		if (write(fd, c, 1) != 1) {
			DEBUG(6, ">>-%s\n", "FAIL");
			DEBUG(1, "ACU write %s\n", "error");
			logent("WRITE ERROR", "ACU FAILED");
			return(FAIL);
		}
		DEBUG(6, (*c >= 040 ? "%c" : "^%c"),(*c >= 040 ? *c : *c+0100));
		c++;
	}
	DEBUG(6, ">>-%s\n", "OK");
	return(SUCCESS);
}


/*
 *  substr(s, l)
 *
 *  Function:	Checks for the presence of the string pointed to by s
 *		somewhere within the string pointed to by l.
 *
 *  Returns:	SUCCESS if found.
 *		FAIL if not found.
 */
substr(s, l)
	char *s;
	register char *l;
{
	int len;

	len = strlen(s);
	while ((l = strchr(l, *s)) != NULL) {
		if (strncmp(s, l, len) == SAME)
			return(SUCCESS);
		l++;
	}
	return(FAIL);
}


/*
 *  alrmint()
 *
 *  Function:	Catches alarm calls (signal 14) and exits.
 *
 *  Returns:	No return.  Exits with status FAIL.
 */
alrmint()
{
	DEBUG(4, "\nTimeout waiting for %s\n", dialing ? "carrier" : "acu");
	logent((dialing ? "WAITING FOR CARRIER" : "WAITING FOR LOCAL ACU"),
		 "TIMEOUT");
	cleanup(FAIL);
}


/*
 *  cleanup(stat)
 *
 *  Function:	Closes device file and exits.
 *
 *  Returns:	No return.  Exits with status stat.
 */
cleanup(stat)
	int stat;
{
	close(fd);
	exit(stat);
}
